# -*- coding: utf-8 -*-

import contextlib
import operator
import os

import six

import pypugjs
from pypugjs.exceptions import CurrentlyNotSupported
from pypugjs.runtime import iteration, escape, open


def process_param(key, value, terse=False):
    if terse:
        if (key == value) or (value is True):
            return key
    if isinstance(value, six.binary_type):
        value = value.decode('utf8')
    return '''%s="%s"''' % (key, value)


TYPE_CODE = {
    'if': operator.truth,
    'unless': operator.not_,
    'elsif': operator.truth,
    'else': lambda v: True,
}


@contextlib.contextmanager
def local_context_manager(compiler, local_context):
    old_local_context = compiler.local_context
    new_local_context = dict(compiler.local_context)
    new_local_context.update(local_context)
    compiler.local_context = new_local_context
    yield
    compiler.local_context = old_local_context


class Compiler(pypugjs.compiler.Compiler):
    global_context = {}
    local_context = {}
    mixins = {}
    useRuntime = True

    def _do_eval(self, value):
        if isinstance(value, six.string_types):
            value = value.encode('utf-8')
        try:
            value = eval(value, self.global_context, self.local_context)
        except Exception:
            return None
        return value

    def _get_value(self, attr):
        value = attr['val']
        if attr['static']:
            return attr['val']
        if isinstance(value, six.string_types):
            return self._do_eval(value)
        else:
            return attr['name']

    def _make_mixin(self, mixin):
        arg_names = [arg.strip() for arg in mixin.args.split(",")]

        def _mixin(self, args):
            if args:
                arg_values = self._do_eval(args)
            else:
                arg_values = []
            if not isinstance(arg_values, (list, tuple)):
                arg_values = [arg_values]
            local_context = dict(zip(arg_names, arg_values))
            with local_context_manager(self, local_context):
                self.indents, old_indent = 0, self.indents
                self.visitBlock(mixin.block)
                self.indents = old_indent

        return _mixin

    def interpolate(self, text, escape=True):
        esc = self.html_escape if escape else lambda x: x
        return self._interpolate(text, lambda x: esc(str(self._do_eval(x))))

    def visitInclude(self, node):
        path = os.path.join(self.options.get("basedir", os.getcwd()), node.path)
        if os.path.exists(path):
            src = open(path, 'r').read()
        elif os.path.exists("%s.pug" % path):
            src = open("%s.pug" % path, 'r').read()
        else:
            raise Exception("Include path doesn't exists")

        parser = pypugjs.parser.Parser(src)
        block = parser.parse()
        self.visit(block)

    def visitExtends(self, node):
        raise CurrentlyNotSupported()

    def visitMixin(self, mixin):
        if mixin.block:
            self.mixins[mixin.name] = self._make_mixin(mixin)
        else:
            self.mixins[mixin.name](self, mixin.args)

    def visitAssignment(self, assignment):
        self.global_context[assignment.name] = self._do_eval(assignment.val)

    def visitConditional(self, conditional):
        if not conditional.sentence:
            value = False
        else:
            value = self._do_eval(conditional.sentence)
        if TYPE_CODE[conditional.type](value):
            self.visit(conditional.block)
        elif conditional.next:
            for item in conditional.next:
                self.visitConditional(item)

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()
            val = self.var_processor(val)
            val = self._do_eval(val)
            if code.escape:
                val = self.html_escape(str(val))
            self.buf.append(val)
        if code.block:
            self.visit(code.block)
        if not code.buffer and not code.block:
            six.exec_(code.val.lstrip(), self.global_context, self.local_context)

    def visitEach(self, each):
        obj = iteration(self._do_eval(each.obj), len(each.keys))
        for item in obj:
            local_context = {}
            if len(each.keys) > 1:
                for (key, value) in zip(each.keys, item):
                    local_context[key] = value
            else:
                local_context[each.keys[0]] = item
            with local_context_manager(self, local_context):
                self.visit(each.block)

    def attributes(self, attrs):
        return " ".join(['''%s="%s"''' % (k, v) for (k, v) in attrs.items()])

    def visitDynamicAttributes(self, attrs):
        classes = []
        params = []
        for attr in attrs:
            if attr['name'] == 'class':
                value = self._get_value(attr)
                if isinstance(value, list):
                    classes.extend(value)
                else:
                    classes.append(value)
            else:
                value = self._get_value(attr)
                if value is True:
                    params.append((attr['name'], True))
                elif value not in (None, False):
                    params.append((attr['name'], escape(value)))
        if classes:
            classes = [six.text_type(c) for c in classes]
            params.append(('class', " ".join(classes)))
        if params:
            self.buf.append(
                " " + " ".join([process_param(k, v, self.terse) for (k, v) in params])
            )


HTMLCompiler = Compiler


def process_pugjs(src, **options):
    parser = pypugjs.parser.Parser(src)
    block = parser.parse()
    compiler = Compiler(block, pretty=True, **options)
    return compiler.compile()



import os

from jinja2.ext import Extension
from jinja2.runtime import Undefined
from markupsafe import Markup

import pypugjs.runtime
from pypugjs import Compiler as _Compiler
from pypugjs.runtime import attrs as _attrs, iteration, open
from pypugjs.utils import process

ATTRS_FUNC = '__pypugjs_attrs'
ITER_FUNC = '__pypugjs_iter'


def attrs(attrs, terse=False):
    return Markup(_attrs(attrs, terse, Undefined))


class Compiler(_Compiler):
    def visitCodeBlock(self, block):
        if self.mixing > 0:
            if self.mixing > 1:
                caller_name = '__pypugjs_caller_%d' % self.mixing
            else:
                caller_name = 'caller'
            self.buffer(
                '{%% if %s %%}%s %s() %s{%% endif %%}'
                % (
                    caller_name,
                    self.variable_start_string,
                    caller_name,
                    self.variable_end_string,
                )
            )
        else:
            self.buffer('{%% block %s %%}' % block.name)
            if block.mode == 'append':
                self.buffer(
                    '%ssuper()%s'
                    % (self.variable_start_string, self.variable_end_string)
                )
            self.visitBlock(block)
            if block.mode == 'prepend':
                self.buffer(
                    '%ssuper()%s'
                    % (self.variable_start_string, self.variable_end_string)
                )
            self.buffer('{% endblock %}')

    def visitMixin(self, mixin):
        self.mixing += 1
        if not mixin.call:
            self.buffer('{%% macro %s(%s) %%}' % (mixin.name, mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('{% endmacro %}')
        elif mixin.block:
            if self.mixing > 1:
                self.buffer('{%% set __pypugjs_caller_%d=caller %%}' % self.mixing)
            self.buffer('{%% call %s(%s) %%}' % (mixin.name, mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('{% endcall %}')
        else:
            self.buffer(
                '%s%s(%s)%s'
                % (
                    self.variable_start_string,
                    mixin.name,
                    mixin.args,
                    self.variable_end_string,
                )
            )
        self.mixing -= 1

    def visitAssignment(self, assignment):
        self.buffer('{%% set %s = %s %%}' % (assignment.name, assignment.val))

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()
            val = self.var_processor(val)
            self.buf.append(
                '%s%s%s%s'
                % (
                    self.variable_start_string,
                    val,
                    '|escape' if code.escape else '',
                    self.variable_end_string,
                )
            )
        else:
            self.buf.append('{%% %s %%}' % code.val)

        if code.block:
            # if not code.buffer: self.buf.append('{')
            self.visit(code.block)
            # if not code.buffer: self.buf.append('}')

            if not code.buffer:
                codeTag = code.val.strip().split(' ', 1)[0]
                if codeTag in self.auto_close_code:
                    self.buf.append('{%% end%s %%}' % codeTag)

    def visitEach(self, each):
        self.buf.append(
            "{%% for %s in %s(%s,%d) %%}"
            % (','.join(each.keys), ITER_FUNC, each.obj, len(each.keys))
        )
        self.visit(each.block)
        self.buf.append('{% endfor %}')

    def visitInclude(self, node):
        path = os.path.join(
            self.options.get("basedir", '.'), self.format_path(node.path)
        )
        if os.path.exists(path):
            src = open(path, 'r').read()
        else:
            raise Exception("Include path doesn't exists ({})".format(path))

        parser = pypugjs.parser.Parser(src)
        block = parser.parse()
        self.visit(block)

    def attributes(self, attrs):
        return "%s%s(%s)%s" % (
            self.variable_start_string,
            ATTRS_FUNC,
            attrs,
            self.variable_end_string,
        )


class PyPugJSExtension(Extension):
    options = {}
    file_extensions = '.pug'

    def __init__(self, environment):
        super(PyPugJSExtension, self).__init__(environment)

        environment.extend(pypugjs=self)

        environment.globals[ATTRS_FUNC] = attrs
        environment.globals[ITER_FUNC] = iteration
        self.variable_start_string = environment.variable_start_string
        self.variable_end_string = environment.variable_end_string
        self.options["variable_start_string"] = environment.variable_start_string
        self.options["variable_end_string"] = environment.variable_end_string

    def preprocess(self, source, name, filename=None):
        loader = self.environment.loader
        try:
            # we're in a Flask app
            loader = loader.app.jinja_loader
        except AttributeError:
            pass
        if hasattr(loader, 'searchpath') and len(loader.searchpath):
            self.options["basedir"] = loader.searchpath[0]

        if not name or (name and not os.path.splitext(name)[1] in self.file_extensions):
            return source
        return process(source, filename=name, compiler=Compiler, **self.options)



from __future__ import absolute_import

from .compiler import Compiler


def register_filter(name=None):
    def decorator(f):
        Compiler.register_filter(name, f)
        return f

    return decorator


@register_filter('cdata')
def cdata_filter(x, y):
    return '<![CDATA[\n%s\n]]>' % x


try:
    import coffeescript

    @register_filter('coffeescript')
    def coffeescript_filter(x, y):
        return '<script>%s</script>' % coffeescript.compile(x)


except ImportError:
    pass

try:
    import markdown

    @register_filter('markdown')
    def markdown_filter(x, y):
        return markdown.markdown(x, output_format='html5')


except ImportError:
    pass



from collections import deque

import six


class Node(object):
    debug = False

    def __str__(self):
        return self.__dict__.__str__()


class BlockComment(Node):
    def __init__(self, val, block, buffer):
        self.block = block
        self.val = val
        self.buffer = buffer


class Block(Node):
    def __init__(self, node=None):
        self.nodes = deque()
        self.debug = False
        if node:
            self.append(node)

    def replace(self, other):
        other.nodes = self.nodes

    def append(self, node):
        return self.nodes.append(node)

    def prepend(self, node):
        return self.nodes.appendleft(node)

    def isEmpty(self):
        return bool(self.nodes)

    def unshift(self, node):
        return self.nodes.unshift(node)


class CodeBlock(Block):
    pass


class Code(Node):
    def __init__(self, val, buffer, escape):
        self.val = val
        self.block = None
        self.buffer = buffer
        self.escape = escape


class Comment(Node):
    def __init__(self, val, buffer):
        self.val = val
        self.buffer = buffer


class Doctype(Node):
    def __init__(self, val):
        self.val = val


class Each(Node):
    def __init__(self, obj, keys, block=None):
        self.obj = obj
        self.keys = keys
        self.block = block


class Assignment(Node):
    def __init__(self, name, val):
        self.name = name
        self.val = val


class Mixin(Node):
    def __init__(self, name, args, block, call):
        self.name = name
        self.args = args
        self.block = block
        self.call = call


class Extends(Node):
    def __init__(self, path):
        self.path = path


class Include(Node):
    def __init__(self, path, extra=None):
        self.path = path
        self.extra = extra


class Conditional(Node):
    may_contain_tags = {
        'if': ['elif', 'else'],
        'for': ['else'],
        'elif': ['elif', 'else'],
        'unless': ['elif', 'else'],
    }

    def __init__(self, type, sentence, block=None):
        self.type = type
        self.sentence = sentence
        self.block = block
        self.next = []

    def can_append(self, type):
        n = (self.next and self.next[-1].type) or self.type
        return type in self.may_contain_tags.get(n, [])

    def append(self, next):
        self.next.append(next)


class Filter(Node):
    def __init__(self, name, block, attrs):
        self.name = name
        self.block = block
        self.attrs = attrs
        self.isASTFilter = isinstance(block, Block)


class Literal(Node):
    def __init__(self, str):
        self.str = str.replace('\\', '\\\\')


class Tag(Node):
    def __init__(self, name, block=None, inline=False, buffer=False):
        self.name = name
        self.textOnly = False
        self.code = None
        self.text = None
        self._attrs = []
        self.inline = inline
        self.block = block or Block()
        self.buffer = buffer

    @classmethod
    def static(self, string, only_remove=False):
        if not isinstance(string, six.string_types) or not string:
            return string
        if string[0] in ('"', "'"):
            if string[0] == string[-1]:
                string = string[1:-1]
            else:
                return string
        if only_remove:
            return string
        return '"%s"' % string

    def setAttribute(self, name, val, static=True):
        self._attrs.append(dict(name=name, val=val, static=static))
        return self

    def removeAttribute(self, name):
        for attr in self._attrs:
            if attr and attr['name'] == name:
                self._attrs.remove(attr)

    def getAttribute(self, name):
        for attr in self._attrs:
            if attr and attr['name'] == name:
                return attr['val']

    @property
    def attrs(self):
        attrs = []
        classes = []
        static_classes = True
        for attr in self._attrs:
            name = attr['name']
            val = attr['val']
            static = attr['static']  # and isinstance(val,six.string_types)
            if static:
                val = self.static(val)
            if val in ("True", "False", "None"):
                val = val == "True"
                static = True
            d = dict(name=name, val=val, static=static)
            if name == 'class':
                static_classes = static_classes and static
                classes.append(d)
            else:
                attrs.append(d)
        if classes:
            if static_classes:
                classes = [
                    dict(
                        name='class',
                        val='"%s"' % ' '.join([a['val'][1:-1] for a in classes]),
                        static=True,
                    )
                ]
            else:
                for attr in classes:
                    attr['static'] = static_classes
        return attrs + classes


class Text(Node):
    parent = None

    def __init__(self, line=None):
        self.nodes = []
        if isinstance(line, six.string_types):
            self.append(line)

    def append(self, node):
        return self.nodes.append(node)


class String(Text):
    def __init__(self, line=None, inline=False):
        super(String, self).__init__(line=line)
        self.inline = inline



from __future__ import absolute_import

import six

from . import nodes
from .lexer import Lexer


class Parser(object):
    def __init__(self, source, filename=None, **options):
        self.input = source
        self.lexer = Lexer(source, **options)
        self.filename = filename
        self.bloks = {}
        self.options = options
        self.contexts = [self]
        self.extending = False
        self._spaces = None

    def context(self, parser):
        if parser:
            self.contexts.append(parser)
        else:
            self.contexts.pop()

    def advance(self):
        return self.lexer.advance()

    def skip(self, n):
        while n > 1:  # > 0?
            self.advance()
            n -= 1

    def peek(self):
        p = self.lookahead(1)
        return p

    def line(self):
        return self.lexer.lineno

    def lookahead(self, n):
        return self.lexer.lookahead(n)

    def parse(self):
        block = nodes.Block()
        block.line = self.line()

        while 'eos' != self.peek().type:
            if 'newline' == self.peek().type:
                self.advance()
            else:
                block.append(self.parseExpr())

        # TODO: this is kinda broken and unused what is it doing?
        # parser = self.extending
        # if parser:
        #     self.context(parser)
        #     ast = parser.parse()
        #     self.context(parser)
        #     return ast

        return block

    def expect(self, string):
        t = self.peek().type
        if t == string:
            return self.advance()
        else:
            raise Exception(
                'expected "%s" but got "%s" in file %s on line %d'
                % (string, t, self.filename, self.line())
            )

    def accept(self, attrs):
        if self.peek().type == attrs:
            return self.advance()

    def parseExpr(self):
        t = self.peek().type
        if 'yield' == t:
            self.advance()
            block = nodes.Block()
            block._yield = True
            return block
        elif t in ('id', 'class'):
            tok = self.advance()
            new_div = self.lexer.tok('tag', 'div')
            new_div.inline_level = tok.inline_level
            self.lexer.stash.append(new_div)
            self.lexer.stash.append(tok)
            return self.parseExpr()

        funcName = 'parse%s' % t.capitalize()
        if hasattr(self, funcName):
            return getattr(self, funcName)()
        else:
            raise Exception(
                'unexpected token "%s" in file %s on line %d'
                % (t, self.filename, self.line())
            )

    def parseString(self):
        tok = self.expect('string')
        node = nodes.String(tok.val, inline=tok.inline_level > 0)
        node.line = self.line()
        return node

    def parseText(self):
        tok = self.expect('text')
        node = nodes.Text(tok.val)
        node.line = self.line()
        return node

    def parseBlockExpansion(self):
        if ':' == self.peek().type:
            self.advance()
            return nodes.Block(self.parseExpr())
        else:
            return self.block()

    def parseAssignment(self):
        tok = self.expect('assignment')
        return nodes.Assignment(tok.name, tok.val)

    def parseCode(self):
        tok = self.expect('code')
        node = nodes.Code(tok.val, tok.buffer, tok.escape)  # tok.escape
        block, i = None, 1
        node.line = self.line()
        while self.lookahead(i) and 'newline' == self.lookahead(i).type:
            i += 1
        block = 'indent' == self.lookahead(i).type
        if block:
            self.skip(i - 1)
            node.block = self.block()
        return node

    def parseComment(self):
        tok = self.expect('comment')

        if 'indent' == self.peek().type:
            node = nodes.BlockComment(tok.val, self.block(), tok.buffer)
        else:
            node = nodes.Comment(tok.val, tok.buffer)

        node.line = self.line()
        return node

    def parseDoctype(self):
        tok = self.expect('doctype')
        node = nodes.Doctype(tok.val)
        node.line = self.line()
        return node

    def parseFilter(self):
        tok = self.expect('filter')
        attrs = self.accept('attrs')
        self.lexer.pipeless = True
        block = self.parseTextBlock()
        self.lexer.pipeless = False

        node = nodes.Filter(tok.val, block, attrs and attrs.attrs)
        node.line = self.line()
        return node

    def parseASTFilter(self):
        tok = self.expect('tag')
        attrs = self.accept('attrs')

        self.expect(':')
        block = self.block()

        node = nodes.Filter(tok.val, block, attrs and attrs.attrs)
        node.line = self.line()
        return node

    def parseEach(self):
        tok = self.expect('each')
        node = nodes.Each(tok.code, tok.keys)
        node.line = self.line()
        node.block = self.block()
        return node

    def parseConditional(self):
        tok = self.expect('conditional')
        node = nodes.Conditional(tok.val, tok.sentence)
        node.line = self.line()
        node.block = self.block()
        while True:
            t = self.peek()
            if 'conditional' == t.type and node.can_append(t.val):
                node.append(self.parseConditional())
            else:
                break
        return node

    def parseExtends(self):
        path = self.expect('extends').val.strip('"\'')
        return nodes.Extends(path)

    def parseCall(self):
        tok = self.expect('call')
        name = tok.val
        args = tok.args
        if args is None:
            args = ""
        block = self.block() if 'indent' == self.peek().type else None
        return nodes.Mixin(name, args, block, True)

    def parseMixin(self):
        tok = self.expect('mixin')
        name = tok.val
        args = tok.args
        if args is None:
            args = ""
        block = self.block() if 'indent' == self.peek().type else None
        return nodes.Mixin(name, args, block, block is None)

    def parseBlock(self):
        block = self.expect('block')
        mode = block.mode
        name = block.val.strip()
        block = (
            self.block(cls=nodes.CodeBlock)
            if 'indent' == self.peek().type
            else nodes.CodeBlock(nodes.Literal(''))
        )
        block.mode = mode
        block.name = name
        return block

    def parseInclude(self):
        path = self.expect('include').val.strip()
        return nodes.Include(path)

    def parseTextBlock(self, tag=None):
        text = nodes.Text()
        text.line = self.line()
        if tag:
            # TODO: needs a test because this line was text.parent == tag
            # its untested and i don't know what it shall do right now
            text.parent = tag
        spaces = self.expect('indent').val
        if not self._spaces:
            self._spaces = spaces
        indent = ' ' * (spaces - self._spaces)
        while 'outdent' != self.peek().type:
            t = self.peek().type
            if 'newline' == t:
                text.append('\n')
                self.advance()
            elif 'indent' == t:
                text.append('\n')
                for node in self.parseTextBlock().nodes:
                    text.append(node)
                text.append('\n')
            else:
                text.append(indent + self.advance().val)

        if spaces == self._spaces:
            self._spaces = None
        self.expect('outdent')
        return text

    def block(self, cls=nodes.Block):
        block = cls()
        block.line = self.line()
        self.expect('indent')
        while 'outdent' != self.peek().type:
            if 'newline' == self.peek().type:
                self.advance()
            else:
                block.append(self.parseExpr())
        self.expect('outdent')
        return block

    def processInline(self, current_tag, current_level):
        next_level = current_level + 1
        while self.peek().inline_level == next_level:
            current_tag.block.append(self.parseExpr())

        if self.peek().inline_level > next_level:
            self.processInline(current_tag, next_level)

    def processTagText(self, tag):
        if self.peek().inline_level < tag.inline_level:
            return

        if not self.lookahead(2).inline_level > tag.inline_level:
            tag.text = self.parseText()
            return

        while (
            self.peek().inline_level == tag.inline_level
            and self.peek().type == 'string'
        ):
            tag.block.append(self.parseExpr())

            if self.peek().inline_level > tag.inline_level:
                self.processInline(tag, tag.inline_level)

    def parseTag(self):
        i = 2
        if 'attrs' == self.lookahead(i).type:
            i += 1

        if ':' == self.lookahead(i).type:
            if 'indent' == self.lookahead(i + 1).type:
                raise Exception(
                    'unexpected token "indent" in file %s on line %d'
                    % (self.filename, self.line())
                )

        tok = self.advance()
        tag = nodes.Tag(tok.val, buffer=tok.val[0] == '#')
        tag.inline_level = tok.inline_level

        tag.line = self.line()

        while True:
            t = self.peek().type
            if t in ('id', 'class'):
                tok = self.advance()
                tag.setAttribute(tok.type, '"%s"' % tok.val, True)
                continue
            elif 'attrs' == t:
                tok = self.advance()
                for n, v in six.iteritems(tok.attrs):
                    tag.setAttribute(n, v, n in tok.static_attrs)
                continue
            else:
                break

        v = self.peek().val
        if '.' == v:
            tag.textOnly = True
            self.advance()
        elif '<' == v:  # For inline elements
            tag.inline = True
            self.advance()

        t = self.peek().type
        if 'code' == t:
            tag.code = self.parseCode()
        elif ':' == t:
            self.advance()
            tag.block = nodes.Block()
            tag.block.append(self.parseExpr())
        elif 'string' == t:
            self.processTagText(tag)
        elif 'text' == t:
            tag.text = self.parseText()

        while 'newline' == self.peek().type:
            self.advance()

        if 'indent' == self.peek().type:
            if tag.textOnly:
                self.lexer.pipeless = True
                tag.block = self.parseTextBlock(tag)
                self.lexer.pipeless = False
            else:
                block = self.block()
                if tag.block:
                    for node in block.nodes:
                        tag.block.append(node)
                else:
                    tag.block = block

        return tag



from pypugjs.ext.mako import preprocessor

try:
    from pyramid_mako import MakoRendererFactory
    from pyramid_mako import parse_options_from_settings
    from pyramid_mako import PkgResourceTemplateLookup

    is_pyramid_mako = True
except ImportError:
    from pyramid import mako_templating

    is_pyramid_mako = False


class PyPugJSRenderer(object):
    """
    The PugJS renderer
    """

    def __init__(self, info):
        info.settings['mako.preprocessor'] = preprocessor
        self.makoRenderer = mako_templating.renderer_factory(info)

    def __call__(self, value, system):
        return self.makoRenderer(value, system)


def add_pugjs_renderer(config, extension, mako_settings_prefix='mako.'):
    renderer_factory = MakoRendererFactory()
    config.add_renderer(extension, renderer_factory)

    def register():
        settings = config.registry.settings
        settings['{0}preprocessor'.format(mako_settings_prefix)] = preprocessor

        opts = parse_options_from_settings(
            settings, mako_settings_prefix, config.maybe_dotted
        )
        lookup = PkgResourceTemplateLookup(**opts)

        renderer_factory.lookup = lookup

    config.action(('pug-renderer', extension), register)


def includeme(config):
    if not is_pyramid_mako:
        # looks broken, but i dont know how to fix
        config.add_renderer(".pug", renderer)  # noqa
    else:
        config.add_directive('add_pugjs_renderer', add_pugjs_renderer)
        config.add_pugjs_renderer('.pug')



from __future__ import print_function

import codecs
import logging
import os
import sys
from optparse import OptionParser

from pypugjs.utils import process


def convert_file():
    support_compilers_list = [
        'django',
        'jinja',
        'underscore',
        'mako',
        'tornado',
        'html',
    ]
    available_compilers = {}
    for i in support_compilers_list:
        try:
            compiler_class = __import__(
                'pypugjs.ext.%s' % i, fromlist=['pypugjs']
            ).Compiler
        except ImportError as e:
            logging.warning(e)
        else:
            available_compilers[i] = compiler_class

    usage = "usage: %prog [options] [file [output]]"
    parser = OptionParser(usage)
    parser.add_option(
        "-o", "--output", dest="output", help="Write output to FILE", metavar="FILE"
    )
    # use a default compiler here to sidestep making a particular
    # compiler absolutely necessary (ex. django)
    default_compiler = sorted(available_compilers.keys())[0]
    parser.add_option(
        "-c",
        "--compiler",
        dest="compiler",
        choices=list(available_compilers.keys()),
        default=default_compiler,
        type="choice",
        help=(
            "COMPILER must be one of %s, default is %s"
            % (', '.join(list(available_compilers.keys())), default_compiler)
        ),
    )
    parser.add_option(
        "-e",
        "--ext",
        dest="extension",
        help="Set import/extends default file extension",
        metavar="FILE",
    )

    options, args = parser.parse_args()

    file_output = options.output or (args[1] if len(args) > 1 else None)
    compiler = options.compiler

    if options.extension:
        extension = '.%s' % options.extension
    elif options.output:
        extension = os.path.splitext(options.output)[1]
    else:
        extension = None

    if compiler in available_compilers:
        import six

        if len(args) >= 1:
            template = codecs.open(args[0], 'r', encoding='utf-8').read()
        elif six.PY3:
            template = sys.stdin.read()
        else:
            template = codecs.getreader('utf-8')(sys.stdin).read()
        output = process(
            template,
            compiler=available_compilers[compiler],
            staticAttrs=True,
            extension=extension,
        )
        if file_output:
            outfile = codecs.open(file_output, 'w', encoding='utf-8')
            outfile.write(output)
        elif six.PY3:
            sys.stdout.write(output)
        else:
            codecs.getwriter('utf-8')(sys.stdout).write(output)
    else:
        raise Exception('You must have %s installed!' % compiler)


if __name__ == '__main__':
    convert_file()



from django.conf import settings
from django.utils.encoding import force_str as to_text
from django.utils.translation import template

from pypugjs import Compiler as _Compiler, register_filter
from pypugjs.exceptions import CurrentlyNotSupported
from pypugjs.utils import process


class Compiler(_Compiler):
    auto_close_code = [
        'autoescape',
        'cache',
        'comment',
        'compress',
        'block',
        'blocktrans',
        'filter',
        'for',
        'if',
        'ifchanged',
        'ifequal',
        'ifnotequal',
        'localize',
        'spaceless',
        'trans',
        'with',
        'verbatim',
    ]
    useRuntime = True

    def __init__(self, node, **options):
        if settings.configured:
            options.update(getattr(settings, 'PYPUGJS', {}))
        super(Compiler, self).__init__(node, **options)

    def visitCodeBlock(self, block):
        self.buffer('{%% block %s %%}' % block.name)
        if block.mode == 'append':
            self.buffer('{{block.super}}')
        self.visitBlock(block)
        if block.mode == 'prepend':
            self.buffer('{{block.super}}')
        self.buffer('{% endblock %}')

    def visitAssignment(self, assignment):
        self.buffer('{%% __pypugjs_set %s = %s %%}' % (assignment.name, assignment.val))

    def visitMixin(self, mixin):
        self.mixing += 1
        if not mixin.call:
            self.buffer('{%% __pypugjs_kwacro %s %s %%}' % (mixin.name, mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('{% end__pypugjs_kwacro %}')
        elif mixin.block:
            raise CurrentlyNotSupported("The mixin blocks are not supported yet.")
        else:
            self.buffer('{%% __pypugjs_usekwacro %s %s %%}' % (mixin.name, mixin.args))
        self.mixing -= 1

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()
            val = self.var_processor(val)
            self.buf.append('{{%s%s}}' % (val, '|force_escape' if code.escape else ''))
        else:
            self.buf.append('{%% %s %%}' % code.val)

        if code.block:
            self.visit(code.block)

            if not code.buffer:
                code_tag = code.val.strip().split(' ', 1)[0]
                if code_tag in self.auto_close_code:
                    self.buf.append('{%% end%s %%}' % code_tag)

    def attributes(self, attrs):
        return "{%% __pypugjs_attrs %s %%}" % attrs


def decorate_templatize(func):
    def templatize(src, origin=None, charset=None):
        src = to_text(src, charset or settings.FILE_CHARSET)
        if origin.endswith(".pug"):
            html = process(src, compiler=Compiler)
        else:
            html = src
        return func(html, origin)

    return templatize


# fix translation for pug templates
def enable_pug_translations():
    template.templatize = decorate_templatize(template.templatize)


try:
    from django.contrib.markup.templatetags.markup import markdown

    @register_filter('markdown')
    def markdown_filter(x, y):
        return markdown(x)


except ImportError:
    pass



from __future__ import absolute_import

from itertools import chain

import six
from six.moves import reduce

import io

import charset_normalizer

try:
    from collections.abc import Mapping as MappingType
except ImportError:
    import UserDict

    MappingType = (UserDict.UserDict, UserDict.DictMixin, dict)


def escape(s):
    """Convert the characters &, <, >, ' and " in string s to HTML-safe
    sequences.  Use this if you need to display text that might contain
    such characters in HTML.  Marks return value as markup string.
    """
    if hasattr(s, '__html__'):
        return s.__html__()
    if isinstance(s, six.binary_type):
        s = six.text_type(str(s), 'utf8')
    elif isinstance(s, six.text_type):
        s = s
    else:
        s = str(s)

    return (
        s.replace('&', '&amp;')
        .replace('>', '&gt;')
        .replace('<', '&lt;')
        .replace("'", '&#39;')
        .replace('"', '&#34;')
    )


def attrs(attrs=None, terse=False, undefined=None):
    if attrs is None:
        attrs = []

    def extract_classes(cls):
        """Recursively extract_class from iterable and mappings"""
        if isinstance(cls, (list, tuple)):
            return tuple(reduce(lambda t1, t2: t1 + t2, map(extract_classes, cls)))
        if isinstance(cls, dict):
            return tuple(
                sorted(dict(filter(lambda x: x[1] and x[1] != undefined, cls.items())))
            )
        return (str(cls),)

    buf = []
    if bool(attrs):
        buf.append(u'')
        for k, v in attrs:
            if undefined is not None and isinstance(v, undefined):
                continue
            if v is not None and (v or type(v) != bool):
                if k == 'class':
                    v = u' '.join(extract_classes(v))
                t = v and type(v) == bool
                if t and not terse:
                    v = k
                buf.append(u'%s' % k if terse and t else u'%s="%s"' % (k, escape(v)))
    return u' '.join(buf)


def is_mapping(value):
    return isinstance(value, MappingType)


def is_iterable(ob):
    if isinstance(ob, six.string_types):
        return False
    try:
        iter(ob)
        return True
    except TypeError:
        return False


def get_cardinality(ob):
    if isinstance(ob, six.string_types):
        return 1
    try:
        return len(ob)
    except TypeError:
        return 1


def iteration(obj, num_keys):
    """
    PugJS iteration supports "for 'value' [, key]?" iteration only.
    PyPugJS has implicitly supported value unpacking instead, without
    the list indexes. Trying to not break existing code, the following
    rules are applied:

      1. If the object is a mapping type, return it as-is, and assume
         the caller has the correct set of keys defined.

      2. If the object's values are iterable (and not string-like):
         a. If the number of keys matches the cardinality of the object's
            values, return the object as-is.
         b. If the number of keys is one more than the cardinality of
            values, return a list of [v(0), v(1), ... v(n), index]

      3. Else the object's values are not iterable, or are string like:
         a. if there's only one key, return the list
         b. otherwise return a list of (value,index) tuples

    """

    # If the object is a mapping type, return it as-is
    if is_mapping(obj):
        return obj

    _marker = []

    iter_obj = iter(obj)
    head = next(iter_obj, _marker)
    iter_obj = chain([head], iter_obj)

    if head is _marker:
        # Empty list
        return []

    if is_iterable(head):
        cardinality = get_cardinality(head)
        if cardinality > 0 and num_keys == cardinality + 1:
            return (tuple(item) + (ix,) for ix, item in enumerate(iter_obj))
        else:
            return iter_obj

    elif num_keys == 2:
        return ((item, ix) for ix, item in enumerate(iter_obj))

    else:
        return iter_obj


def open(
    file, mode='r', buffering=-1, encoding=None, errors=None, newline=None, closefd=True
):
    if encoding is None:
        charset_match = charset_normalizer.from_path(file).best()
        encoding = charset_match and charset_match.encoding

    decoded = io.open(
        file,
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        closefd=closefd,
    )

    return decoded
